%Event-Trggered control of a Linear system with bounded data rate
%instantaneous communication

%Author: Pavan Tallapragada

%function inst_comm()
function [ TH,X,TJ,XJ,OP ] = inst_comm(pbar)

%%%%%%%%% Params %%%%%%%%%%%%
%%

%{
A = [[1.38, -0.2077 6.715 -5.676]; ...
     [-0.5814 -4.29 0 0.675]; ...
     [1.067 4.273 -6.654 5.893]; ...
     [0.048 4.273 1.343 -2.104] ];
B = [ 0 0; 5.679 0; 1.136 -3.146; 1.136 0];
K = -place(A,B,[-3 -3 -2 -2])
%}
%
A = [1 -2; 1 4];
B = [0; 1];
K = -place(A,B,[-1 -2]);
%

[n,n] = size(A);
[n,m] = size(B);

Q = eye(n);

Abar = A+B*K;
P = lyap(Abar', Q);

Lm_Q = min(eig(Q));
LM_P = max(eig(P));
Lm_P = min(eig(P));

beta = 0.8*(Lm_Q/LM_P);
a = 1.2;
%sigma = 1;

%uniform bound on the Euclidean norm of the disturbance
nu = 0;

%steady state value of Vd
V0 = 0;

%load a sequence of random integers between 1 and pbar (not used currently,
%simply transmit the prescribed minimum number of bits)

Rp = 1;
%Rp = 1 + floor( pbar.*rand(50,1) );
%{
if pbar == 12
    load('rand_p_pbar12.mat', 'Rp');
end

if pbar == 20
    load('rand_p_pbar20.mat', 'Rp');
end
%}

opt = odeset('reltol',1e-12,'abstol',1e-12,'events',@comm_event);
T0 = 0;
Tf = 40;

%initial conditions
%xp0 = [5, -3, -4, 4]';
%xh0 = [4.9, -2.9, -4.1, 4.1]';

xp0 = [6, -4]';
xh0 = [0, 0]';

de0 = 2*norm(xp0 - xh0, Inf);
Vd0 = 1.1*xp0'*P*xp0;
p_Lbnd0 = 0;
p0 = 0;
k0 = 0;
t_jmp0 = 0;

x0 = [xp0; xh0; Vd0; de0; p_Lbnd0; p0; k0; t_jmp0];

TH = [ ];
X = [ ];
TJ = [ ];
XJ = [ ];
OP = [ ];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

w = (Lm_Q/LM_P) - beta;
W = (Lm_Q/LM_P) - a*beta;
c = (W*sqrt(Lm_P))/(2*sqrt(n)*norm(P*B*K));

if W <= 0
    disp('check W');
    return;
end

alpha_A = norm(A);
theta = alpha_A + beta/2;

c1 = 0;
c2 = 0;

%---------------------------

Params.T0 = T0;
Params.Tf = Tf;
Params.Rp = Rp;
Params.n = n;
Params.m = m;
Params.A = A;
Params.B = B;
Params.K = K;
Params.P = P;
Params.Q = Q;
Params.V0 = V0;
Params.nu = nu;
Params.alpha_A = alpha_A;
Params.beta = beta;
Params.w = w;
Params.W = W;
Params.theta = theta;
Params.pbar = pbar;
Params.c = c;
Params.c1 = c1;
Params.c2 = c2;
Params.de0 = de0;
Params.Vd0 = Vd0;

%---------------------------

G1 = Gamma1(1, 1, Params)

T = 0.5*G1; %The parameter in the rho function

%coefficient of (1 - b) in the definition of the rho function
c3 = ( w + theta )/( W * ( exp( (w + theta)*T ) - 1 ) );

%---------------------------

Params.c3 = c3;
Params.G1 = G1;
Params.T = T;

%---------------------------

%G2 = Gamma2(1, 1, 1, Params)
G2_unif = Gamma2_uniform(1, 2^pbar, Params)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

V_t0 = xp0'*P*xp0;
b0 = V_t0/Vd0;

eps0 = de0/(c*sqrt(Vd0));
hch0 = eps0 / rho(b0, c3);
disp('hch0 / (2^(pbar))');
disp( hch0 / (2^(pbar)) )

if hch0 > 2^(pbar)
    disp('check and change params');
    return;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%aggregate state of the hybrid system
x0 = [xp0; xh0; Vd0; de0; p_Lbnd0; p0; k0; t_jmp0];

TermId = [ ];

q = 1;
Pat = cell(q,1);

Pat{1} = 1:2*n+6;

[ TH,X,TJ,XJ,OP ]=hybrid_sys( @OL_control, ...
                              @fin_dat_comm, @comm_event, ...
                              @outputs_fun, TermId, Pat, T0, Tf, x0, ...
                              opt, Params );

%plot_fun( TH, X, TJ, XJ, OP, Params );

%-------------------------------------------------------
function dx = OL_control(t,x,Params)

n = Params.n;
A = Params.A;
B = Params.B;
K = Params.K;
beta = Params.beta;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last jump time

dxp = A*xp + B*K*xh;
dxh = (A+B*K)*xh;
dVd = -beta*Vd;
dde = 0;
dp_Lbnd = 0;
dp = 0;
dk = 0;
dt_jmp = 0;

dx = [dxp; dxh; dVd; dde; dp_Lbnd; dp; dk; dt_jmp];

%-------------------------------------------------------
function x_new = fin_dat_comm(th, x, Params, jump_flag)

n = Params.n;
A = Params.A;
B = Params.B;
K = Params.K;
c = Params.c;
beta = Params.beta;
pbar = Params.pbar;
c3 = Params.c3;
Rp = Params.Rp;
P = Params.P;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last jump time

%de just before jumping
de_ = norm(expm(A*(th(1)-t_jmp)),Inf)*de;

V = xp'*P*xp;
b = V/Vd;

k = k + 1;
p_Lbnd = ceil(log2( de_/( c*rho(b,c3)*sqrt(Vd) ) ));
%because of numerical errors ceil may make p_Lbnd larger than pbar. It may
%occur when h2 part is triggered in the triggering rule, in which case
%p_Lbnd should just be pbar. so do the following.
p_Lbnd = min(p_Lbnd, pbar);

p = p_Lbnd;     %choose the minimum prescribed number of bits
%p = max([p_Lbnd, Rp(k)]);

de = de_;

for i = 1 : p
    de = de / 2;
    xh = xh - de*((xp-xh)<=0) + de*((xp-xh)>0);
end

t_jmp = th(1);

x_new = [ xp; xh; Vd; de; p_Lbnd; p; k; t_jmp ];

%---------------------------------------------------------
function [value,isterminal,direction] = comm_event(t,x,Params)

n = Params.n;
A = Params.A;
B = Params.B;
K = Params.K;
P = Params.P;
Q = Params.Q;
c = Params.c;
c3 = Params.c3;
beta = Params.beta;
pbar = Params.pbar;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de_last = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last jump time

xe = xp - xh;
V = xp'*P*xp;

hV = V - Vd;
dhV = - xp'*Q*xp - 2*xp'*P*B*K*xe + beta*Vd;
term1 = min(hV, dhV);

b = V/Vd;
de_cur = norm(expm(A*(t-t_jmp)),Inf)*de_last;
hch = de_cur / (c*rho(b,c3)*sqrt(Vd));

term2 = hch - 2^pbar;

value = max(term1, term2);

isterminal = ones(1,1);   %Stop the integration
direction = ones(1,1);    %Increasing directions,
                            %(as initially the error is zero)

%-------------------------------------------------------
function OP = outputs_fun( TH, X, TJ, XJ, Params )

n = Params.n;
A = Params.A;
B = Params.B;
K = Params.K;
P = Params.P;
Q = Params.Q;
c = Params.c;
beta = Params.beta;
pbar = Params.pbar;
de0 = Params.de0;
Vd0 = Params.Vd0;
Tsim = Params.Tf - Params.T0;

Xp = X(:,1:n);
Xh = X(:,n+1:2*n);
Xe = Xp - Xh;
De_last = X(:,2*n+2);
T_jmp = X(:,2*n+6);

q = 1;

Nevents = zeros(q,1);
Avg_int = -1*ones(q,1);
mITJ = -1*ones(q,1);

for i = 1:q
    Nevents(i) = length( TJ{i} );
    if Nevents(i) >= 2
        Avg_int(i) = (TJ{i}(end) - TJ{i}(1))/(Nevents(i)-1);
        mITJ(i) = min(TJ{i}(2:end) - TJ{i}(1:end-1));        
    end
end

disp('Tsim');
disp(Tsim);
disp('# updates');
disp( Nevents' );
disp('Avg interval.');
disp(Avg_int');
disp('min interval');
disp(mITJ');

%Set outputs to be returned
OP.n = n;
OP.V = sum(Xp'.*(P*Xp'))';            %Lyapunov function evolution
OP.Vd = X(:,2*n+1);
OP.hV = OP.V - OP.Vd;
OP.p_Lbnd = X(:,2*n+3);
OP.p = X(:,2*n+4);

OP.de = De_last;
for i = 1 : length(T_jmp)
    OP.de(i) = norm(expm(A*(TH(i,1)-T_jmp(i))),Inf).*De_last(i);
end

y = n*(norm(A, inf) + beta/2)*log2(exp(1))*TJ{1} ...
    + n*log2(de0/(c*sqrt(Vd0))) + n;
OP.y = y;

cP = sqrt(det(inv(P)))*(pi^(n/2))/(gamma(n/2  + 1));
z = (trace(A) + n*beta/2)*log2(exp(1))*TJ{1} ...
    + log2( ((2*de0)^n)/( cP*Vd0^(n/2) ) );
OP.z = z;

OP.eps = OP.de ./ (c*sqrt(OP.Vd));
OP.pbar = pbar;

OP.Nevents = Nevents;
OP.Avg_int = Avg_int;
OP.mITJ = mITJ;

%-------------------------------------------------------
function OP = plot_fun( TH, X, TJ, XJ, OP, Params )

n = Params.n;

%%
figure; h1 = gcf();
plot(TH(:,1),OP.V,'b',TH(:,1),OP.Vd,'b--','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
legend('$V$', '$V_d$');
set(legend(), 'interpreter','latex','fontsize', 20)

figure; h2 = gcf();
plot(TJ{1},n*cumsum([XJ{1}(:,2*n+4)]),'b','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('Total # bits transmitted','fontsize',20);
%legend('$\bar{p} = 12$', '$\bar{p} = 20$');
%set(legend(), 'interpreter','latex','fontsize', 20)

figure; h3 = gcf();
stem(TJ{1},n*[XJ{1}(:,2*n+4)],'bo','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('# bits transmitted','fontsize',20);

%%
%{
figure(h1)
saveas(gcf,'./fig_plots/pbar_20_V','fig');
set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
print -depsc ./plots/pbar_20_V
close

figure(h4)
saveas(gcf,'./fig_plots/bit_rate_compare','fig');
set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
print -depsc ./plots/bit_rate_compare
close

figure(h5)
saveas(gcf,'./fig_plots/bit_profile','fig');
set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
print -depsc ./plots/bit_profile
close
%}