%Event-Trggered control of a Linear system with bounded data rate
%non-instantaneous communication with zero disturbance

%Author: Pavan Tallapragada

%function non_inst_comm_disturb()
function [ TH,X,TJ,XJ,OP ] = non_inst_comm(pbar, varargin)

if nargin >= 2
    %transmit max number of bits for the first vargin(1) number of samples
    tx_max_bits_upto_sample = varargin{1};
    to_plot = 0;
else
    tx_max_bits_upto_sample = 0;
    to_plot = 1;
end

%%%%%%%%% Params %%%%%%%%%%%%
%%

%{
A = [[1.38, -0.2077 6.715 -5.676]; ...
     [-0.5814 -4.29 0 0.675]; ...
     [1.067 4.273 -6.654 5.893]; ...
     [0.048 4.273 1.343 -2.104] ];
B = [ 0 0; 5.679 0; 1.136 -3.146; 1.136 0];
K = -place(A,B,[-3 -3 -2 -2])
%}
%
A = [1 -2; 1 4];
B = [0; 1];
K = -place(A,B,[-1 -2]);
%

[n,n] = size(A);
[n,m] = size(B);

Q = eye(n);

Abar = A+B*K;
P = lyap(Abar', Q);

Lm_Q = min(eig(Q));
LM_P = max(eig(P));
Lm_P = min(eig(P));

beta = 0.8*(Lm_Q/LM_P);
a = 1.2;
sigma = 0.9;

%uniform bound on the Euclidean norm of the disturbance
nu = 0;

%steady state value of Vd
V0 = ( 2*norm(P)*nu / ( sigma*(a-1)*beta*sqrt(Lm_P) ) )^2

opt = odeset('reltol',1e-12,'abstol',1e-12,'events',@comm_event);
T0 = 0;
Tf = 40;

%initial conditions
%xp0 = [5, -3, -4, 4]';
%xh0 = [4.9, -2.9, -4.1, 4.1]';

xp0 = [6, -4]';
xh0 = [0, 0]';

de0 = 2*norm(xp0 - xh0, Inf);
Vd0 = 1.1*xp0'*P*xp0;
p_Lbnd0 = 0;
p0 = 0;
k0 = 0;
t_jmp0 = 0;
r_jmp0 = 0;
z0 = zeros(n, 1);
zD0 = zeros(n, 1);
del0 = de0;
intrans0 = 0;    %0: no message in transit, 1: message in transit

%aggregate state of the hybrid system
x0 = [xp0; xh0; Vd0; de0; p_Lbnd0; p0; k0; t_jmp0; r_jmp0; z0; zD0; ...
      del0; intrans0];

TH = [ ];
X = [ ];
TJ = [ ];
XJ = [ ];
OP = [ ];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

w = (Lm_Q/LM_P) - beta;
W = (Lm_Q/LM_P) - a*beta;
c = (W*sqrt(Lm_P))/(2*sqrt(n)*norm(P*B*K));

if W <= 0
    disp('check W');
    return;
end

alpha_A = norm(A);
theta = alpha_A + beta/2;

%c1 = ( 2*norm(P)*nu ) / ( sqrt(Lm_P*V0) );
%c2 = ( W*nu ) / ( c*alpha_A*sqrt(V0) );
c1 = 0;
c2 = 0;

%---------------------------

Params.tx_max_bits_upto_sample = tx_max_bits_upto_sample;
Params.T0 = T0;
Params.Tf = Tf;
Params.n = n;
Params.m = m;
Params.A = A;
Params.B = B;
Params.K = K;
Params.P = P;
Params.Q = Q;
Params.V0 = V0;
Params.nu = nu;
Params.alpha_A = alpha_A;
Params.beta = beta;
Params.w = w;
Params.W = W;
Params.theta = theta;
Params.pbar = pbar;
Params.c = c;
Params.c1 = c1;
Params.c2 = c2;

%---------------------------

G1 = Gamma1(1, 1, Params)

T = 0.5*G1; %The parameter in the rho function

%coefficient of (1 - b) in the definition of the rho function
c3 = ( w + theta )/( W * ( exp( (w + theta)*T ) - 1 ) );

%---------------------------

Params.c3 = c3;
Params.G1 = G1;
Params.T = T;

%---------------------------

%G2 = Gamma2(1, 1, 1, Params)
G2_unif = Gamma2_uniform( 1/(2^pbar), 1, Params )

%TM = 0.8*min([G1, T, G2_unif])
TM = 1.1*1e-3

%---------------------------

Params.TM = TM;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

V_t0 = xp0'*P*xp0;
b0 = V_t0/Vd0;

eps0 = de0/(c*sqrt(Vd0));
psi0 = eps0 / (2^pbar);

if b_tilde(TM, b0, eps0, Params) > 1
    disp('check params, b_tilde(TM, b0, eps0, Params)');
    disp( b_tilde(TM, b0, eps0, Params) )
    return;
end

if hch_bar(TM, b0, eps0, psi0, Params) > 1
    disp('check params, hch_bar(TM, b0, eps0, psi0, Params)');
    disp( hch_bar(TM, b0, eps0, psi0, Params) )    
    return;
end

%return;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
TermId = [ ];

q = 2;
Pat = cell(q,1);

Pat{1} = 1:4*n+9;
Pat{2} = 1:4*n+9;

[ TH,X,TJ,XJ,OP ]=hybrid_sys( @OL_control, ...
                              @fin_dat_comm_jmp, @comm_event, ...
                              @outputs_fun, TermId, Pat, T0, Tf, x0, ...
                              opt, Params );

%to_plot = 1;
if to_plot == 1
    plot_fun( TH, X, TJ, XJ, OP, Params );
end

%-------------------------------------------------------

function btild_val = b_tilde(t, b0, eps0, Par)

w = Par.w;
W = Par.W;
alpha_A = Par.alpha_A;
theta = Par.theta;
c1 = Par.c1;
c2 = Par.c2;


f1 = b0 + W*eps0*( exp((w+theta)*t) - 1 )/(w+theta) + ...
     (c1 - c2)*( exp(w*t) - 1 )/w + ...
     c2*( exp((w+alpha_A)*t) - 1 )/(w+alpha_A);

f2 = exp(w*t);

btild_val = f1/f2;

%-------------------------------------------------------

function hch_bar_val = hch_bar(t, b0, eps0, psi0, Par)

A = Par.A;
alpha_A = Par.alpha_A;
beta = Par.beta;
nu = Par.nu;
V0 = Par.V0;
c = Par.c;
c3 = Par.c3;

b = b_tilde(t, b0, eps0, Par);

num = norm(expm(A*t), Inf) * exp(beta*t/2) * psi0 ;% ...
      %+ ( nu / ( c*alpha_A*sqrt(V0) ) ) * ( exp(alpha_A*t) - 1 );

hch_bar_val = num / rho(b, c3);

%-------------------------------------------------------
function dx = OL_control(t,x,Params)

n = Params.n;
A = Params.A;
B = Params.B;
K = Params.K;
beta = Params.beta;
V0 = Params.V0;
nu = Params.nu;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last transmit time
r_jmp = x(2*n+7);   %last reception time
z = x(2*n+7+1:3*n+7);
zD = x(3*n+7+1:4*n+7);
del = x(4*n+8);
intrans = x(4*n+9);

%disturbance v(t)
v1 = 0;
v2 = 0;
v = [v1; v2];

dxp = A*xp + B*K*xh + v;
dxh = (A+B*K)*xh;
dVd = -beta*(Vd - V0);
dde = 0;
dp_Lbnd = 0;
dp = 0;
dk = 0;
dt_jmp = 0;
dr_jmp = 0;

dz = zeros(n,1);
dzD = zeros(n,1);
ddel = 0;
dintrans = 0;

dx = [dxp; dxh; dVd; dde; dp_Lbnd; dp; dk; dt_jmp; dr_jmp; dz; dzD; ...
      ddel; dintrans];

%-------------------------------------------------------
function x_new = fin_dat_comm_jmp(th, x, Params, jump_flag)

n = Params.n;
A = Params.A;
B = Params.B;
K = Params.K;
c = Params.c;
c3 = Params.c3;
pbar = Params.pbar;
P = Params.P;
nu = Params.nu;
alpha_A = Params.alpha_A;
TM = Params.TM;
tx_max_bits_upto_sample = Params.tx_max_bits_upto_sample;

Abar = A + B*K;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last transmit time
r_jmp = x(2*n+7);   %last reception time
z = x(2*n+7+1:3*n+7);
zD = x(3*n+7+1:4*n+7);
del = x(4*n+8);
intrans = x(4*n+9);

if jump_flag(1) >= 0    %means transmit jump
    
    k = k + 1;
    
    %de just before jumping
    de_ = norm(expm(A*(th(1)-t_jmp)),Inf)*del + (nu / alpha_A)*( ...         %check/correct this
          exp(alpha_A*(th(1) - t_jmp)) - 1 );

    V = xp'*P*xp;
    b = V/Vd;

    eps = de_ / ( c*sqrt(Vd) );
    
    %hch_cur = eps / rho(b, c3);
    
    for jj = 1:pbar
        psi = eps/(2^jj);
        hch_bar_val = hch_bar(TM, b, eps, psi, Params);
        if hch_bar_val <= 0 
            disp('hch_bar negative where it should not be');
        end
        if hch_bar_val <= 1
            break;
        end
    end
        
    p_Lbnd = jj;
    p = p_Lbnd;
    
    if k <= tx_max_bits_upto_sample
        p = pbar;
    end
    
    del = de_;
    
    z = xh;
    zD = xh;

    for i = 1 : p
        del = del / 2;
        zD = zD - del*((xp-zD)<=0) + del*((xp-zD)>0);
    end
    
    t_jmp = th(1);
    intrans = 1;    %message in transit
end

if jump_flag(2) >= 0    %means reception jump
    Delta = th(1) - t_jmp;
    xh = expm(Abar*Delta)*z + expm(A*Delta)*(zD - z);
    
    de = norm(expm(A*(Delta)),Inf)*del + ...
         nu*( expm(alpha_A*Delta) - 1)/alpha_A ;
    
    r_jmp = th(1);
    intrans = 0;    %all messages recieved
end

x_new = [xp; xh; Vd; de; p_Lbnd; p; k; t_jmp; r_jmp; z; zD; del; intrans];

%---------------------------------------------------------
function [value,isterminal,direction] = comm_event(t,x,Params)

n = Params.n;
A = Params.A;
P = Params.P;
nu = Params.nu;
alpha_A = Params.alpha_A;
c = Params.c;

pbar = Params.pbar;
TM = Params.TM;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last transmit time
r_jmp = x(2*n+7);   %last reception time
z = x(2*n+7+1:3*n+7);
zD = x(3*n+7+1:4*n+7);
del = x(4*n+8);
intrans = x(4*n+9);

xe = xp - xh;
V = xp'*P*xp;

b = V/Vd;
de_cur = norm(expm(A*(t-t_jmp)),Inf)*del + (nu / alpha_A)*( ...         %check/correct this
         exp(alpha_A*(t - t_jmp)) - 1 );
eps = de_cur / (c*sqrt(Vd));
psi = eps / ( 2^pbar );

term1 = b_tilde(TM, b, eps, Params) -1;
term2 = hch_bar(TM, b, eps, psi, Params) - 1;

if intrans == 0
    value(1) = max(term1, term2);
    value(2) = - TM;
else
    value(1) = -sign(max(term1, term2))*max(term1, term2);
    value(2) = (t - t_jmp) - TM;
end

value = value(:);

if value(1) >= 0
    1;
end

isterminal = ones(2,1);   %Stop the integration
direction = ones(2,1);    %Increasing directions,
                            %(as initially the error is zero)

%-------------------------------------------------------
function OP = outputs_fun( TH, X, TJ, XJ, Params )

n = Params.n;
A = Params.A;
P = Params.P;
c = Params.c;
nu = Params.nu;
alpha_A = Params.alpha_A;
pbar = Params.pbar;
Tsim = Params.Tf - Params.T0;

Xp = X(:,1:n);
Xh = X(:,n+1:2*n);
Xe = Xp - Xh;
T_jmp = X(:,2*n+6);
R_jmp = X(:,2*n+7);
Del_last = X(:,4*n+8);

q = 2;

Nevents = zeros(q,1);
Avg_int = zeros(q,1);
mITJ = -1*ones(q,1);

for i = 1:q
    Nevents(i) = length( TJ{i} );
    Avg_int(i) = Tsim/Nevents(i);
    if Nevents(i) >= 2
        mITJ(i) = min(TJ{i}(2:end) - TJ{i}(1:end-1));        
    end
end

disp('Tsim');
disp(Tsim);
disp('# updates');
disp( Nevents' );
disp('Avg interval.');
disp(Avg_int');
disp('min interval');
disp(mITJ');

%Set outputs to be returned
OP.n = n;
OP.V = sum(Xp'.*(P*Xp'))';            %Lyapunov function evolution
OP.Vd = X(:,2*n+1);
OP.hV = OP.V - OP.Vd;
OP.p_Lbnd = X(:,2*n+3);
OP.p = X(:,2*n+4);

OP.InfNXe = max(abs(Xe'))';

OP.de = Del_last;
for i = 1 : length(T_jmp)
    OP.de(i) = norm(expm(A*(TH(i,1)-T_jmp(i))),Inf).*Del_last(i) + ...
               (nu / alpha_A)*( exp(alpha_A*(TH(i,1)-T_jmp(i))) - 1 ) ;
end

OP.eps = OP.de ./ (c*sqrt(OP.Vd));
OP.pbar = pbar;

OP.Nevents = Nevents;
OP.Avg_int = Avg_int;
OP.mITJ = mITJ;

%-------------------------------------------------------
function OP = plot_fun( TH, X, TJ, XJ, OP, Params )

n = Params.n;

%%
figure; h1 = gcf();
plot(TH(:,1),OP.V,'b',TH(:,1),OP.Vd,'b--','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
legend('$V$', '$V_d$');
set(legend(), 'interpreter','latex','fontsize', 20)

figure; h2 = gcf();
plot(TJ{1},n*cumsum([XJ{1}(:,2*n+4)]),'b','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('Total # bits transmitted','fontsize',20);
%legend('$\bar{p} = 12$', '$\bar{p} = 20$');
%set(legend(), 'interpreter','latex','fontsize', 20)

figure; h3 = gcf();
stem(TJ{1},n*[XJ{1}(:,2*n+4)],'bo','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('# bits transmitted','fontsize',20);

%%
%{
figure(h1)
saveas(gcf,'./fig_plots/non_inst_pbar_20_V','fig');
set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
print -depsc ./plots/non_inst_pbar_20_V
close

figure(h2)
saveas(gcf,'./fig_plots/non_inst_pbar_20_cum_bit','fig');
set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
print -depsc ./plots/non_inst_pbar_20_cum_bit
close

figure(h3)
saveas(gcf,'./fig_plots/non_inst_pbar_20_bit_profile','fig');
set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
print -depsc ./plots/non_inst_pbar_20_bit_profile
close
%}